from django.template import engines
from django.template.utils import InvalidTemplateEngineError
from threading import Thread
from typing import List
from ssr.backends import javascript

threads = []  # type: List[Thread]

try:
    engine = engines['javascript']  # type: javascript.Components
except InvalidTemplateEngineError:
    raise EnvironmentError(
        'Server side rendering improperly configured. Did you forget '
        'to include the template engine in your Django settings?'
    )

engine.setup()
production_mode = engine.production_mode


def use_server() -> None:
    threads.append(Thread(target=engine.server.run))


def use_builders() -> None:
    for bundler in engine.bundlers:
        threads.append(Thread(target=bundler.build))


def use_watchers() -> None:
    for bundler in engine.bundlers:
        threads.append(Thread(target=bundler.watch))


def run() -> None:
    for thread in threads:
        thread.start()
    for thread in threads:
        thread.join()



from os.path import relpath, splitext, basename, join, dirname, exists
from urllib.parse import urljoin
from collections import namedtuple
from django.template import TemplateSyntaxError
from django.http import HttpRequest
from ssr.settings import ENTRIES_DIR, BUNDLES_DIR, CACHE_DIR, SOCKETS_DIR
from ssr.server import Server


class Template:
    Env = namedtuple('Env', [
        'entry',
        'out_file',
        'out_dir',
        'script_relpath',
        'socket',
        'cache_dir'
    ])

    def __init__(self, path: str, root: str, production_mode: bool,
                 server: Server, server_script: str, client_script: str,
                 build_hash: str, static_url: str, dist_dir: str) -> None:
        self.render_server = server
        self.relpath = relpath(path, root)

        root_relpath = splitext(self.relpath)[0]
        socket_relpath = root_relpath + '-bundler.sock'

        out_relpath = root_relpath + '.js'
        out_file = basename(out_relpath)
        if production_mode:
            hashed_out_relpath = root_relpath
            hashed_out_relpath += '-' + build_hash + '.js'
        else:
            hashed_out_relpath = out_relpath
        hashed_out_file = basename(hashed_out_relpath)

        out_dir = dirname(self.relpath)
        self.url = urljoin(static_url, out_dir)
        self.component_relpath = relpath(path, ENTRIES_DIR)

        self.server = self.Env(
            entry=join(ENTRIES_DIR, 'server.js'),
            out_file=out_file,
            out_dir=join(BUNDLES_DIR, out_dir),
            script_relpath=relpath(server_script, ENTRIES_DIR),
            socket=join(SOCKETS_DIR, 'client', socket_relpath),
            cache_dir=join(CACHE_DIR, 'server')
        )
        self.client = self.Env(
            entry=join(ENTRIES_DIR, 'client.js'),
            out_file=hashed_out_file,
            out_dir=join(dist_dir, out_dir),
            script_relpath=relpath(client_script, ENTRIES_DIR),
            socket=join(SOCKETS_DIR, 'server', socket_relpath),
            cache_dir=join(CACHE_DIR, 'client')
        )

        self.script = urljoin(static_url, hashed_out_relpath)

        stylesheet_relpath = splitext(out_relpath)[0] + '.css'
        hashed_stylesheet_relpath = splitext(hashed_out_relpath)[0] + '.css'

        if exists(join(BUNDLES_DIR, stylesheet_relpath)):
            self.stylesheet = urljoin(static_url, hashed_stylesheet_relpath)
        else:
            self.stylesheet = ''

    def render(self, context: dict = None, request: HttpRequest = None) -> str:
        template_path = join(self.server.out_dir, self.server.out_file)
        try:
            return self.render_server.render(
                template_path, self.script, self.stylesheet, context)
        except Exception as exception:
            raise TemplateSyntaxError(exception)



import json
from os import makedirs, remove, setsid, environ
from os.path import exists, dirname
from urllib.parse import quote_plus, urlencode
from subprocess import Popen, PIPE
from threading import Thread
from time import sleep
from typing import Callable, Dict
from requests import codes
from requests_unixsocket import Session
from ssr.settings import BUNDLER
from ssr.template import Template
from ssr.utils import wait_for_signal, read_output


class Bundler:
    session = Session()

    def __init__(self, template: Template,
                 env: Dict[str, str], cache: bool) -> None:
        self.template = template
        self.env = env
        self.cache = cache

    def get_server_env(self, watch: bool) -> Dict[str, str]:
        return {
            'SOCKET': self.template.server.socket,
            'COMPONENT': self.template.component_relpath,
            'SCRIPT': self.template.server.script_relpath,
            'PARCEL_OPTIONS': json.dumps({
                'entry': self.template.server.entry,
                'config': {
                    'outDir': self.template.server.out_dir,
                    'outFile': self.template.server.out_file,
                    'cache': self.cache,
                    'cacheDir': self.template.server.cache_dir,
                    'watch': watch,
                    'sourceMaps': False
                }
            })
        }

    def get_client_env(self, watch: bool) -> Dict[str, str]:
        return {
            'SOCKET': self.template.client.socket,
            'COMPONENT': self.template.component_relpath,
            'SCRIPT': self.template.client.script_relpath,
            'PARCEL_OPTIONS': json.dumps({
                'entry': self.template.client.entry,
                'config': {
                    'outDir': self.template.client.out_dir,
                    'outFile': self.template.client.out_file,
                    'cache': self.cache,
                    'cacheDir': self.template.client.cache_dir,
                    'watch': watch,
                    'publicUrl': self.template.url
                }
            })
        }

    def build(self) -> None:
        server_env = self.get_server_env(False)
        client_env = self.get_client_env(False)
        server_thread = Thread(target=self.bundle, args=[server_env, False])
        client_thread = Thread(target=self.bundle, args=[client_env, False])
        self.run(server_thread, client_thread)

    def watch(self) -> None:
        server_env = self.get_server_env(True)
        client_env = self.get_client_env(True)
        server_thread = Thread(target=self.prepare, args=[
            self.template.server.socket, lambda: self.bundle(server_env, True)
        ])
        client_thread = Thread(target=self.prepare, args=[
            self.template.client.socket, lambda: self.bundle(client_env, True)
        ])
        self.run(server_thread, client_thread)

    def run(self, server_thread: Thread, client_thread: Thread) -> None:
        server_thread.start()
        client_thread.start()
        server_thread.join()
        client_thread.join()

    def prepare(self, socket: str, watcher: Callable[[], None]) -> None:
        makedirs(dirname(socket), exist_ok=True)
        if exists(socket):
            if self.reconnect(socket):
                return
            remove(socket)
        watcher()
        self.connect(socket)

    def poll(self, socket: str) -> None:
        url = self.get_url(socket)
        while True:
            response = self.session.get(url)
            message = response.content.strip(b' ')
            if message:
                print(message.decode('utf-8')[:-1])
            else:
                sleep(0.1)

    def get_url(self, socket: str) -> str:
        return 'http+unix://' + quote_plus(socket) + '?' + urlencode({
            'pid': self.env['DJANGO_PID']
        })

    def connect(self, socket: str) -> None:
        Thread(target=self.poll, daemon=True, args=[socket]).start()

    def reconnect(self, socket: str) -> bool:
        try:
            url = self.get_url(socket)
            response = self.session.get(url)
            connection_exists = response.status_code == codes.ok
            if connection_exists:
                self.connect(socket)
            return connection_exists
        except:
            return False

    def bundle(self, env: Dict[str, str], watch: bool) -> None:
        process = Popen([
            'node', BUNDLER
        ], stdout=PIPE, stderr=PIPE, preexec_fn=setsid, env={
            **environ,
            **self.env,
            **env
        })
        if watch:
            stdout_thread = Thread(target=wait_for_signal, daemon=True, args=[
                process.stdout, self.env['SIGNAL']
            ])
            stderr_thread = Thread(target=wait_for_signal, daemon=True, args=[
                process.stderr, self.env['SIGNAL']
            ])
            stdout_thread.start()
            stderr_thread.start()
            stdout_thread.join()
            stderr_thread.join()
        else:
            read_output(process)



from os.path import join
from platform import system
from tempfile import gettempdir
from django.conf import settings

BASE_DIR = join(settings.BASE_DIR, '.ssr')
TEMP_DIR = '/tmp' if system() == 'Darwin' else gettempdir()
SCRIPTS_DIR = join(BASE_DIR, 'scripts')
ENTRIES_DIR = join(SCRIPTS_DIR, 'entries')
CACHE_DIR = join(BASE_DIR, 'cache')
BUNDLES_DIR = join(BASE_DIR, 'bundles')
STATIC_DIR = join(BASE_DIR, 'static')
SOCKETS_DIR = join(TEMP_DIR, 'ssr')
HASH_FILE = join(BASE_DIR, 'hash.txt')
SERVER = join(SCRIPTS_DIR, 'server.mjs')
BUNDLER = join(SCRIPTS_DIR, 'bundler.mjs')



import json
from json.encoder import JSONEncoder
from os import makedirs, remove, setsid, environ
from os.path import join, dirname, exists
from urllib.parse import quote_plus, urlencode
from threading import Thread
from subprocess import Popen, PIPE
from requests import codes
from requests_unixsocket import Session
from typing import Dict
from ssr.settings import SERVER, SOCKETS_DIR
from ssr.utils import wait_for_signal


class Server:
    session = Session()
    socket = join(SOCKETS_DIR, 'renderer.sock')

    def __init__(self, env: Dict[str, str],
                 json_encoder: JSONEncoder = None) -> None:
        self.env = env
        self.json_encoder = json_encoder

    def get_url(self, pathname: str = '', params: dict = {}) -> str:
        host = quote_plus(self.socket)
        querystring = urlencode(params)
        return 'http+unix://' + host + pathname + '?' + querystring

    @property
    def exists(self) -> bool:
        url = self.get_url('', {
            'pid': self.env['DJANGO_PID']
        })
        try:
            response = self.session.get(url)
            if response.status_code == codes.ok:
                return True
        except:
            pass
        return False

    def run(self) -> None:
        if self.exists:
            return

        makedirs(dirname(self.socket), exist_ok=True)
        if exists(self.socket):
            remove(self.socket)

        process = Popen([
            'node', SERVER
        ], stdout=PIPE, stderr=PIPE, preexec_fn=setsid, env={
            **environ,
            **self.env,
            'SOCKET': self.socket,
        })
        stdout_thread = Thread(target=wait_for_signal, args=[
            process.stdout, self.env['SIGNAL']
        ])
        stderr_thread = Thread(target=wait_for_signal, args=[
            process.stderr, self.env['SIGNAL']
        ])
        stdout_thread.start()
        stderr_thread.start()
        stdout_thread.join()
        stderr_thread.join()

    def render(self, bundle: str, script: str, stylesheet: str,
               props: dict = None) -> str:
        url = self.get_url('/render', {
            'bundle': bundle,
            'props': json.dumps(props, cls=self.json_encoder),
            'script': script,
            'stylesheet': stylesheet,
        })
        response = self.session.get(url)
        if response.status_code == codes.ok:
            return response.text
        else:
            raise EnvironmentError(response.text)



from os import remove
from os.path import exists
from shutil import rmtree
from django.core.management.base import BaseCommand, CommandError
from ssr.settings import HASH_FILE, BUNDLES_DIR, STATIC_DIR


class Command(BaseCommand):
    help = 'Bundle JavaScript templates using Parcel bundler'

    def handle(self, *args, **options):
        for build_dir in (BUNDLES_DIR, STATIC_DIR):
            if exists(build_dir):
                rmtree(build_dir, ignore_errors=True)
        if exists(HASH_FILE):
            remove(HASH_FILE)  # Generate new hash for cache busting
        from ssr import worker
        worker.use_builders()
        try:
            worker.run()
        except Exception as exception:
            raise CommandError(exception)



from os import getpid, makedirs
from os.path import join, dirname, abspath, exists
from shutil import rmtree, copytree
from urllib.parse import urljoin
from glob import glob
from uuid import uuid1
from django.template.backends.base import BaseEngine
from django.template import TemplateDoesNotExist
from django.utils.module_loading import import_string
from django.core.serializers.json import DjangoJSONEncoder
from django.conf import settings
from typing import Dict, List
from ssr.settings import HASH_FILE, SCRIPTS_DIR, BASE_DIR, STATIC_DIR
from ssr.server import Server
from ssr.bundler import Bundler
from ssr.template import Template


class Components(BaseEngine):
    app_dirname = 'bundles'
    templates = {}  # type: Dict[str, Template]

    def __init__(self, params: dict) -> None:
        params = params.copy()
        self.options = params.pop('OPTIONS').copy()
        super().__init__(params)

    def setup(self) -> None:
        makedirs(BASE_DIR, exist_ok=True)

        if exists(SCRIPTS_DIR):
            rmtree(SCRIPTS_DIR, ignore_errors=True)
        scripts_dir = join(dirname(dirname(abspath(__file__))), 'scripts')
        copytree(scripts_dir, SCRIPTS_DIR)

        if exists(HASH_FILE):
            with open(HASH_FILE, 'r') as hash_file:
                build_hash = hash_file.read()
        else:
            build_hash = str(uuid1())
            with open(HASH_FILE, 'w') as hash_file:
                hash_file.write(build_hash)

        env = {
            'NODE_ENV': 'development' if settings.DEBUG else 'production',
            'NODE_OPTIONS': '-r esm',
            'WORKER_TTL': '1000'
        }  # Dict[str, str]

        if 'env' in self.options:
            env.update(self.options['env'])

        env.update({
            'WORKER_TTL': str(env['WORKER_TTL']),
            'SIGNAL': build_hash,
            'DJANGO_PID': str(getpid())
        })

        self.production_mode = env['NODE_ENV'] == 'production'

        if 'json_encoder' in self.options:
            json_encoder = import_string(self.options['json_encoder'])
        else:
            json_encoder = DjangoJSONEncoder

        self.server = Server(env, json_encoder)

        if 'output_dirname' in self.options:
            output_dirname = self.options['output_dirname']
        else:
            output_dirname = 'dist/'

        static_url = urljoin(settings.STATIC_URL, output_dirname)
        dist_dir = join(STATIC_DIR, output_dirname)

        scripts = {
            'server': join(SCRIPTS_DIR, 'react', 'server.js'),
            'client': join(SCRIPTS_DIR, 'react', 'client.js'),
        }
        if 'scripts' in self.options:
            scripts.update(self.options['scripts'])

        if 'extensions' in self.options:
            extensions = self.options['extensions']
        else:
            extensions = ['js', 'jsx', 'ts', 'tsx']

        if 'cache' in self.options:
            cache = self.options['cache']
        else:
            cache = True

        self.bundlers = []  # type: List[Bundler]

        for template_dir in self.template_dirs:
            for extension in extensions:
                pattern = join(template_dir, '**', '*.' + extension)
                paths = glob(pattern, recursive=True)
                for path in paths:
                    template = Template(
                        path=path,
                        root=template_dir,
                        production_mode=env['NODE_ENV'] == 'production',
                        static_url=static_url,
                        server_script=scripts['server'],
                        client_script=scripts['client'],
                        dist_dir=dist_dir,
                        server=self.server,
                        build_hash=build_hash
                    )
                    self.templates[template.relpath] = template
                    bundler = Bundler(template, env, cache)
                    self.bundlers.append(bundler)

    def get_template(self, template_name: str) -> Template:
        if template_name in self.templates:
            return self.templates[template_name]
        else:
            raise TemplateDoesNotExist('Component not found', backend=self)
